/******************************************************************************

   KGI - S3 86c32x (ViRGE family) chipset driver

   Copyright (C) 1997, 1998  Stefan Mars	[mars@lysator.liu.se]
             (C) 1998, 1999  Jos Hulzink        [josh@stack.nl]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************

   This code was derived from the following source of information:
   [1]	S3 ViRGE Data Book, (c) August 1996 S3 Incorporated

******************************************************************************
*/

#include <kgi/maintainers.h>
#include <asm/mtrr.h>
#include <linux/config.h>

#define	MAINTAINER	"Jos Hulzink (josh@stack.nl)"
#define	DRIVER_NAME	"S3 ViRGE family chipset driver"
#define	DRIVER_REV	"$Revision: 1.35 $"

#define DEBUG_LEVEL	0

#define	S3_86C325
#define	S3_ViRGE
#define ViRGE
#define	HARDWARE_TEXT_CURSOR	0

#include "S3/virge.h"

#define	IO_VGA_BASE	0x03C0
#define	IO_VGA_SIZE	0x20
#define	IO_ACCEL_BASE	0x02E8
#define	IO_ACCEL_SIZE	4

#define	TEXT_MEMORY_SIZE	(64 KB)

#include <kgi/system.h>
#include <kgi/module.h>
#include <linux/pci.h>

#define	IO_VGA			(private->io_vga_region.baseptr-IO_VGA_BASE)
#define	IO_ACCEL		(private->io_accel_region.baseptr-IO_ACCEL_BASE)
#define	MEM_LAW			(private->mem_law_region.baseptr)

#define	MEM_ACCEL		(private->mem_accel)
#define	MEM_VGA			(private->mem_vga)

#define	IO_VGA_REGION		(private->io_vga_region)
#define	IO_ACCEL_REGION		(private->io_accel_region)
#define	MEM_ACCEL_REGION	(private->mem_accel_region)
#define	MEM_LAW_REGION		(private->mem_law_region)

#define	CHIPSET_INFO		(private->chipset_info)

#define	TEXTMODE_MMIO_REGIONS	NULL
#define	GRAPHMODE_MMIO_REGIONS	NULL

#define	XVIRTUAL	(mode->request.virt.x)
#define	YVIRTUAL	(mode->request.virt.y)
#define XVISIBLE (mode->request.visible.x)
#define YVISIBLE (mode->request.visible.y)

#define	NoSEQRegs	0x1A
#define	NoCRTRegs	0x80
#define	NoGRCRegs	0x09
#define	NoATCRegs	0x15

#include "s3_mmio.inc"
#include "s3_text16.inc"

enum tcardtype { none = 0, virge, vx, dxgx, gx2, mx, mxp, mxpmv } cardtype;

  /* pciVirge : pointer to PCI info of detected ViRGE */
  struct pci_dev *pciVirge = NULL;

/* Chipset info */
static struct kgi_chipset s3_ViRGE_info =
{
	"S3 Incorporated",		/* Manufacturer			*/
	"ViRGE",			/* Model			*/
	GM_TEXT16,			/* Available display modes	*/
	0,				/* Flags			*/
	{ 2048, 2048 },			/* Maximum resolution		*/
	NULL,				/* Private data hook		*/
	0,				/* Available RAM (autodetected)	*/
	{ 10000000, 135000000 }		/* DCLK limits			*/
};  

/* This table holds the default vaules for the ViRGE VGA+ registers.	*/
/* A copy of it will be filled in as appropriate during mode		*/
/* programming in kgim_chipset_set_mode().				*/
static struct statetable default_state =
{
        0x2F,							/* MISC	*/
	0x00,							/* FCTL	*/
	{
		0x03, 0x00, 0x0F, 0x20, 0x02, 0x00, 0x00, 0x00, /* SEQ */
		0x00, 0x00, 0xC0, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x41, 0x2D, 0x00, 0x00, 0x00, 0x80, 0x00, 0x00,
		0x40, 0x00
	},
	{
		0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 0x10,	/* CRT	*/
		0x00, 0x40, 0x00, 0x1F, 0x00, 0x00, 0xFF, 0xFF,
		0x00, 0x00, 0x00, 0x00, 0x1F, 0x00, 0x00, 0xE3,	/* 0x1x */
		0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x2x */
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x4D, 0x00, 0x22, 0x10, 0x00, 0x00, 0x19,	/* 0x3x */
		0x48, 0xA5, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x4x */
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x38, 0x00, 0x00, 0x00,	/* 0x5x */
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x6x */
		0x0C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x1C,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x7x */
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	},
	{
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* GRC	*/
		0xFF
	},
	{
		0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,	/* ATC	*/
		0x08, 0x39, 0x3A, 0x3B, 0x3C, 0x3D, 0x3E, 0x30,
		0x84, 0x00, 0x0F, 0x00, 0x00
	}
};

/* VGA I/O register space */
static const struct io_region s3_ViRGE_io_vga = 
{
	IO_NULL,			/* IO region start (virtual)	*/
	IO_VGA_BASE,			/* IO region start (physical)	*/
	IO_VGA_SIZE,			/* IO region size		*/
	0xFFFF,				/* Decode all address bits	*/
	"S3 VGA chipset IO"		/* IO region name		*/
};

/* S3 accelerator I/O register space */
static const struct io_region s3_ViRGE_io_accel = 
{
	IO_NULL,			/* IO region start (virtual)	*/
	IO_ACCEL_BASE,			/* IO region start (physical)	*/
	IO_ACCEL_SIZE,			/* IO region size		*/
	0x03FF,				/* Decode 10 address bits only	*/
	"S3 accelerator I/O"		/* IO region name		*/
};

/* S3 MMIO (old style) register space */
static const struct mem_region s3_ViRGE_mem_accel = 
{
	MEM_NULL,			/* Mem region start (virtual)	*/
	0xA0000,			/* Mem region start (physical)	*/
	0x10000,			/* Mem region size		*/
	-1,				/* Decode all address bits	*/
	"S3 accelerator mem"		/* Mem region name		*/
};

/* ViRGE linear address window space */
static const struct mem_region s3_ViRGE_mem_law = 
{
	MEM_NULL,			/* Mem region start (virtual)	*/
	0,				/* Mem region start (physical)	*/
	0x04000000,			/* Mem region size		*/
	-1,				/* Decode all address bits	*/
	"ViRGE LAW"			/* Mem region name		*/ 
};

void s3_ViRGE_default_interrupt (struct kgi_display *dpy);

/* Information used by the Interrupthandler */
struct {
  kgiu32 enabled; /* 1 = enabled, 0 = disabled */
  kgiu32 bitmap; /* Copy of MM8504 */;
  void (* VerticalSync)(struct kgi_display *);
  void (* EngineDone)(struct kgi_display *);
  void (* CommandFifoOverflow)(struct kgi_display *);
  void (* CommandFifoEmpty)(struct kgi_display *);
  void (* HostDmaDone)(struct kgi_display *);
  void (* CommandDmaDone)(struct kgi_display *);
  void (* Fifo3dEmpty)(struct kgi_display *);
} s3_ViRGE_interrupts = { 
  0,
  0x0000000,
  s3_ViRGE_default_interrupt,
  s3_ViRGE_default_interrupt,
  s3_ViRGE_default_interrupt,
  s3_ViRGE_default_interrupt,
  s3_ViRGE_default_interrupt,
  s3_ViRGE_default_interrupt,
  s3_ViRGE_default_interrupt
};

/* Waring: do not compile this "inline" ! Pointer to procedure needed */
void s3_ViRGE_default_interrupt (struct kgi_display *dpy) {
  return;
}

void s3_ViRGE_interrupt_handler (int irq, void *d, struct pt_regs *regs) {
  struct kgi_display *dpy=d;
  kgiu32 status;

  GET_PRIVATE;

  /* Get interrupt status and clear pending interrupts immediately */
  status = mem_in32(MEM_VGA + 0X0504);
  mem_out32(0x0000007F, MEM_VGA + 0X0504);
  mem_out32(s3_ViRGE_interrupts.bitmap, MEM_VGA + 0X0504);
  CRT_out8(dpy,CRT_in8(dpy,0x11)&~0x30,0x11);  
  if (s3_ViRGE_interrupts.bitmap & 0x00000100)
    CRT_out8(dpy,CRT_in8(dpy,0x11)|0x10,0x11);     

  /* Now act on pending interrupts. No check is made if there is an actual
   * handler, for this takes time. If an unexpected interrupt comes in,
   * it is handled by a default handler (this should never happen :)
   */
   
  if (status & 0x0001) (*(s3_ViRGE_interrupts.VerticalSync)) (dpy);
  if (status & 0x0002) (*(s3_ViRGE_interrupts.EngineDone)) (dpy);
  if (status & 0x0004) (*(s3_ViRGE_interrupts.CommandFifoOverflow)) (dpy);
  if (status & 0x0008) (*(s3_ViRGE_interrupts.CommandFifoEmpty)) (dpy);
  if (status & 0x0010) (*(s3_ViRGE_interrupts.HostDmaDone)) (dpy);
  if (status & 0x0020) (*(s3_ViRGE_interrupts.CommandDmaDone)) (dpy);
  if (status & 0x0040) (*(s3_ViRGE_interrupts.Fifo3dEmpty)) (dpy);
}

int s3_ViRGE_enable_interrupts (struct kgi_display *dpy) {
  GET_PRIVATE;
  kgiu8 crt11;
  
  IS_CRITICAL;
  
  if (pciVirge == NULL) return -E(CHIPSET,UNKNOWN);
  if (pciVirge->irq >= 0x10 || pciVirge->irq ==0) return -E(CHIPSET,UNKNOWN);
  if (request_irq (pciVirge->irq, s3_ViRGE_interrupt_handler, 
  		   SA_SHIRQ, "ViRGE", dpy)) {
    DEBUG1 ("Unable to attach IRQ handler");
    s3_ViRGE_interrupts.enabled = 0;
    return -E(CHIPSET,UNKNOWN);
  }
  
  ENTER_CRITICAL;
  /* Enable interrupt generation */
  CRT_out8(dpy, 0x10, 0x32); /* Interrupt generation enabled */  

  /* Clear vertical retrace interrupt and enable it. */
  crt11 = CRT_in8(dpy,0x11);
  CRT_out8(dpy,crt11 & ~0x30,0x11);  
  CRT_out8(dpy,crt11 | 0x10,0x11);  
  
  /* Clear all pending enhanced interrupts and enable them */
  mem_out32(0x0000007F, MEM_VGA + 0X0504);
  mem_out32(s3_ViRGE_interrupts.bitmap, MEM_VGA + 0X0504);
  LEAVE_CRITICAL;
  
  /* Set flag */
  s3_ViRGE_interrupts.enabled = 1;
  
  return -EOK;    
}

int s3_ViRGE_disable_interrupts (struct kgi_display *dpy) {
  GET_PRIVATE;
  IS_CRITICAL;
  
  ENTER_CRITICAL;
  /* Switch off ViRGE interrupt generator */
  CRT_out8(dpy, 0x00, 0x32);  
  /* Disable all enhanced interrupts. */
  mem_out32(0x00000000, MEM_VGA + 0X0504);
  LEAVE_CRITICAL;

  /* Detach the interrupt handler */
  if (pciVirge->irq >= 0x10 || pciVirge->irq == 0) return EOK;
  free_irq (pciVirge->irq,dpy);

  s3_ViRGE_interrupts.enabled = 0;
  return EOK;
}

/* s3_ViRGE_PCI_detect:
 * Detects a ViRGE by scanning the list of PCI devices
 *
 * known bugs: * PCI allows an user to put multiple ViRGEs in one computer.
 *               This code only recognizes one.
 *             * If there are more videocards, the ViRGE has to claim
 *             	 private IO access while not in MMIO mode 
 */

struct {
  kgiu16 vendor,device;
  enum tcardtype type;
  char *name;
} Pci_ViRGE_List[] = {
  { PCI_VENDOR_ID_S3, PCI_DEVICE_ID_S3_ViRGE, virge, "S3 ViRGE" },
  { PCI_VENDOR_ID_S3, PCI_DEVICE_ID_S3_ViRGE_DXGX, dxgx, "S3 ViRGE GX / DX" },
  { PCI_VENDOR_ID_S3, PCI_DEVICE_ID_S3_ViRGE_VX, vx, "S3 ViRGE VX" },
  { PCI_VENDOR_ID_S3, PCI_DEVICE_ID_S3_ViRGE_GX2, gx2, "S3 ViRGE GX2" },
  { PCI_VENDOR_ID_S3, PCI_DEVICE_ID_S3_ViRGE_MX, mx, "S3 ViRGE MX" },
  { PCI_VENDOR_ID_S3, PCI_DEVICE_ID_S3_ViRGE_MXP, mxp, "S3 ViRGE MXP" },
  { PCI_VENDOR_ID_S3, PCI_DEVICE_ID_S3_ViRGE_MXPMV, mxpmv, "S3 ViRGE MXPMV" },
  { 0, 0, none, "" }
};  

enum tcardtype s3_ViRGE_PCI_detect (struct kgi_display *dpy) {

  kgiu16 i,cmd = 0;

	cardtype=none;  
	for (i=0;cardtype==none && Pci_ViRGE_List[i].vendor!=0 ;i++) {
	  if ((pciVirge=pci_find_device(Pci_ViRGE_List[i].vendor,
	  			      Pci_ViRGE_List[i].device,pciVirge))) 
	    cardtype=Pci_ViRGE_List[i].type;
	}
	if (cardtype != none) {
	  strcpy (s3_ViRGE_info.model,Pci_ViRGE_List[--i].name);
	  NOTICE ("ViRGE: PCI device detected: %s",Pci_ViRGE_List[i].name);
	  DEBUG1 ("ViRGE: Bus: %d Slot: %d IRQ: %d",pciVirge->bus->number,
	  			PCI_SLOT(pciVirge->devfn),pciVirge->irq);	
	  /* Enable PCI device */
	  cmd = pcicfg_in16((pciVirge->bus->number<<24)
	  		   |(pciVirge->devfn<<16) 
	  		   |PCI_COMMAND);
	  if ((cmd & (PCI_COMMAND_IO | PCI_COMMAND_MEMORY)) == PCI_COMMAND_MEMORY) {
	    NOTICE ("ViRGE: Chip in MMIO mode. Is there another driver ?");
	    return none;
	  }
	} 
  return cardtype;
}

/*
** Helper functions.
**
** This section contains:
**
**	void s3_defaults_ViRGE(struct kgi_display *dpy) for setting some
**	register default values that should be read from the chip as the
**	power-up defaults are unknown
**
**	static int s3_detect_ViRGE(struct kgi_display *dpy) is the main
**	detection routine. It will handle chipdetection and call all other
**	helper functions.
**
**	void s3_mmio_only_ViRGE(struct kgi_display *dpy) sets up the ViRGE
**	for MMIO only mode. THIS IS BUGGED AND DISABLED CURRENTLY
**
*/

/* Card detection and setup. The section below contains all code that is
 * needed to detect a S3 ViRGE (both PCI and IO methods) and initialize this
 * driver
 */

/* s3_defaults_ViRGE:
 * Messy code that gets some values form the ViRGE and does a lot of things
 * that don't belong here.
 * 
 * known bugs: none
 */
void s3_defaults_ViRGE(struct kgi_display *dpy)
{
  kgiu32 cursorbase;
  
	GET_PRIVATE;

	       	#define COPYDEFAULT(X) (default_state.CRT[X] = CRT_in8(dpy, X))

		/* COPYDEFAULT(0x60); 
		 * COPYDEFAULT(0x62);
		 */
		 
		COPYDEFAULT(0x68); /* Power on strapping */
		COPYDEFAULT(0x6F); /* Power on strapping */
	#undef COPYDEFAULT
	
	if (cardtype == vx) {
	  default_state.CRT[0x63]= 0x09;
	  default_state.CRT[0x66]= 0x90;
	} else {
	  default_state.CRT[0x63]= 0x00;
	  default_state.CRT[0x66]= 0x89;
	}
		
	SEQ_out8 (dpy, 0x60, 0x0a); /* RAS1 + PD Tristate + 3 MCLK Write */
	SEQ_out8 (dpy, 0x00, 0x0d); /* ViRGE syncs the screen, no features*/
		
	/*CRT_out8(dpy, 0x10, 0x34);  Enable FIFO register */

	if (cardtype == dxgx) {
	  if (CRT_in8(dpy, 0x6f) & 0x01) { /* ViRGE GX */
	    CRT_out8(dpy, (CRT_in8(dpy, 0x36) & 0xF3) | 0x04, 0x36); /* SDRAM*/
	  } else { /* ViRGE DX */
	    CRT_out8(dpy, (CRT_in8(dpy, 0x36) & 0xF3) | 0x00, 0x36); /* 1-cycle EDO */
	  }
	} else {
	  CRT_out8(dpy, (CRT_in8(dpy, 0x36) & 0xF3) | 0x08, 0x36); /* 2-cycle EDO */
	}
	CRT_out8(dpy, 0x1B, 0x37); /* Enable ViRGE, internal clocks */
	
	/* Hardware Graphics Cursor */
	CRT_out8 (dpy, 0x00, 0x45); /* Disable hardware cursor */
	CRT_in8  (dpy, 0x45); /* Reset HGC Color Stack Pointer */
	CRT_out8 (dpy, 0xff, 0x4A); /* Foreground red value */
	CRT_out8 (dpy, 0xff, 0x4A); /* Foreground green value */
	CRT_out8 (dpy, 0xff, 0x4A); /* Foreground blue value */
	CRT_in8  (dpy, 0x45); /* Reset HGC Color Stack Pointer */
	CRT_out8 (dpy, 0x00, 0x4B); /* Background red value */
	CRT_out8 (dpy, 0x00, 0x4B); /* Background green value */
	CRT_out8 (dpy, 0x00, 0x4B); /* Background blue value */
  	CRT_out8 (dpy, 0x00, 0x4e);/* Display complete cursor (x) */
  	CRT_out8 (dpy, 0x00, 0x4f);/* Display complete cursor (y) */
	cursorbase=(CHIPSET_INFO.ramsize >> 10)-1;
	CRT_out8 (dpy, cursorbase >> 8, 0x4c);
	CRT_out8 (dpy, cursorbase & 0xff, 0x4d);
	CRT_out8 (dpy, 0x10, 0x55); /* X11 mode, normal Ramdac operation */
	
	CRT_out8(dpy, 0x00, 0x56); /* ViRGE syncs itself */
	#ifdef BIG_ENDIAN
	  CRT_out8 (dpy, 0x01, 0x54);
	  CRT_out8 (dpy, CR61_SWAP_BYTE_IN_DWORD, 0x61);
	#else
	  CRT_out8 (dpy, 0x00, 0x54);
	  CRT_out8 (dpy, CR61_NO_SWAP, 0x61);
	#endif
	CRT_out8(dpy, 0x00, 0x65); /* Enable Ramdac MMIO access. XFree says
				      this value should be 0, not 1 ! */
	
	private->lram.next	= NULL;
	private->lram.name	= "video ram";
	private->lram.type	= MMAP_TYPE_MMIO|MMAP_PUBLIC | MMAP_FRAMEBUFFER;
	private->lram.prot	= MMIO_RDWR | MMIO_SHARED;
	private->lram.psize	= private->lram.size = CHIPSET_INFO.ramsize;
	private->lram.pstart	= MEM_LAW_REGION.base;

	/* Setting up the pointer */
	private->ptr_and = 0x00FF;
	private->ptr_xor = 0x6F00;
	private->cur_and = 0x00FF;
	private->cur_xor = 0x7F00;

	return;
}

/* s3_ViRGE_mmio_only:
 * Shuts down all IO-port activity of the ViRGE
 */
void s3_ViRGE_mmio_only(struct kgi_display *dpy) {
	GET_PRIVATE;
	IS_CRITICAL;
	ENTER_CRITICAL;
       	SEQ_out8(dpy, SR09_MMIO_ONLY, 0x09);
        default_state.SEQ[0x09] |= SR09_MMIO_ONLY;
        mem_out8(PCI_COMMAND_MEMORY, MEM_VGA + 4);
        LEAVE_CRITICAL;
       	io_free_region(& IO_ACCEL_REGION);
       	io_free_region(& IO_VGA_REGION);
}

/* s3_ViRGE_setup_memory:
 * Detects physical amount of memory, sets up timing, prepares for MMIO mode.
 */

int s3_ViRGE_setup_memory (struct kgi_display *dpy) {
        GET_PRIVATE;
	kgiu16 pcireg12;
	kgiu8 cr58;
	mem_addr base0 = 0;
	
	/* Maybe we still need this to override the base0 address ? */
	/*struct kgim_options_misc *misc = KGIM_OPTIONS(dpy, misc);
	 */
	 
	/* Figure out the MMIO base address
	 */
        pcireg12 = pcicfg_in16((pciVirge->bus->number<<24)
	  		   |(pciVirge->devfn<<16) 
	  		   |0x12);
       	base0 = pcireg12 << 16;
       	MEM_LAW_REGION.size = 64 MB;
       	MEM_LAW_REGION.base = base0 & ~(MEM_LAW_REGION.size - 1);
       	
       	if(!MEM_LAW_REGION.base || mem_check_region(&MEM_LAW_REGION)) 
       		return -EBUSY;
       	if (!(MEM_LAW = mem_claim_region(& MEM_LAW_REGION))) 
       		return -E(CHIPSET, UNKNOWN);

       	NOTICE("ViRGE: LAW address: %.8x", MEM_LAW_REGION.base);

        CHIPSET_INFO.ramsize = 1024 KB;
	if ((cardtype != vx) && (cardtype != gx2) && (cardtype !=mx)) { 
          CHIPSET_INFO.ramsize = 1024 KB;
       	  switch (CRT_in8(dpy, 0x36) & CR36_MEMSIZE_MASK) {
		case CR36_4MB:
		  CHIPSET_INFO.ramsize <<= 1;
       		case CR36_2MB:
       		  CHIPSET_INFO.ramsize <<= 1;
       		case CR36_1MB:
		  NOTICE("ViRGE: Detected %d KB of video memory",
		         CHIPSET_INFO.ramsize >> 10);
		  break;
       		default:
		  NOTICE("ViRGE: Unknown amount of memory. Assuming 1 MB");
		  break;
       	  } /* switch */
       	} else if (cardtype == vx) {
          CHIPSET_INFO.ramsize = 
            2 MB*(1+((CRT_in8(dpy, 0x36) & 0x60) >> 5));
	  NOTICE("ViRGE: VX: Detected %d KB of video memory",
		         CHIPSET_INFO.ramsize >> 10);
	  switch (CRT_in8 (dpy,0x37) & 0x60) {
	    case 1:
	      CHIPSET_INFO.ramsize -= 2 MB;
	    case 2:
	      CHIPSET_INFO.ramsize -= 2 MB;
       	  } /* switch */
       	  NOTICE("ViRGE: VX: On-screen memory: %d", 
		         CHIPSET_INFO.ramsize >> 10);
       	} else {
	  /* Gx2 or Mx chipset */
          CHIPSET_INFO.ramsize = 2 MB;
       	  switch ((CRT_in8(dpy, 0x36) & 0xC0) >> 6) {
       		case 3:
       		  CHIPSET_INFO.ramsize <<= 1;
		case 1:
		  NOTICE("ViRGE: GX2 / MX: Detected %d KB of video memory",
		         CHIPSET_INFO.ramsize >> 10);
		  break;
       		default:
		  NOTICE("ViRGE: GX2 / MX: Unknown amount of memory. Assuming 2 MB");
		  break;
       	  } /* switch */
	  
       	}
	
	/* Enable new MMIO mode 
	 */
	CRT_out8 (dpy, pcireg12 >> 8, 0x59);   /* some DX forget this */ 
	CRT_out8 (dpy, pcireg12 & 0xff, 0x5A); /* Should be done in hardware */
	cr58= (CRT_in8 (dpy, 0x58) & 0x80)|
	      (CHIPSET_INFO.ramsize == 1 MB ? 0x01 :
	      (CHIPSET_INFO.ramsize == 2 MB ? 0x02 : 0x03)) | 0x10 |
	      (cardtype == vx ? 0x40 : 0);
	CRT_out8 (dpy, cr58, 0x58);
       	CRT_out8 (dpy, 0x08, 0x53);
       	CRT_out8 (dpy, 0x8C, 0x31);

	/* Warning: Below the next line, all IO is done in memory space */

	MEM_VGA	= MEM_LAW + 0x1008000;
	MEM_ACCEL = MEM_LAW + 0x1000000;
	
	/* Now MEM_VGA is defined, all IO is done by MMIO
	 * All IO locks can be removed.
	 */
	s3_ViRGE_mmio_only (dpy);

       	
	#ifdef CONFIG_MTRR
	  /* ViRGE cards are buggy as hell concerning MMIO addressing.
	   * Enabling MTRR immediately crashes my system.
  	   */ 
	  /*mtrr_add (MEM_LAW_REGION.base, MEM_LAW_REGION.size, MTRR_TYPE_UNCACHABLE, 1);
	   *DEBUG1("ViRGE: MTRR registered.");
	   */
	#endif
	 
	/* Set up Memory timings
	*SEQ_out8(dpy, default_state.SEQ[0x10], 0x10);
	*SEQ_out8(dpy, default_state.SEQ[0x11], 0x11);
	*SEQ_out8(dpy, default_state.SEQ[0x15]| 0x01 , 0x15);
	*SEQ_out8(dpy, default_state.SEQ[0x15], 0x15);
	*/
	dpy->card->chipset = & CHIPSET_INFO;
	return EOK;
}

int s3_ViRGE_IO_claim (struct kgi_display *dpy) {
  GET_PRIVATE; 
       	if (io_check_region(& IO_VGA_REGION) ||
	    io_check_region(& IO_ACCEL_REGION)) {
       		ERROR("Chipset: io_vga or io_accel region busy.");
       		return -EBUSY;
       	}
	
       	if (!(io_claim_region(& IO_VGA_REGION) ||
	      io_claim_region(& IO_ACCEL_REGION))) {
       		ERROR("Chipset: Failed to claim io_vga or io_accel region.");
       		return -EBUSY;
       	}
  return EOK;
}

int s3_ViRGE_IO_disable (struct kgi_display *dpy){
  GET_PRIVATE;
  io_free_region(&IO_ACCEL_REGION);
  io_free_region(&IO_VGA_REGION);
  return 0;
}
  
static int s3_ViRGE_init(struct kgi_display *dpy)
{
  GET_PRIVATE;

  private->mem_vga	= MEM_NULL;
  private->mem_accel	= MEM_NULL;
  
  /* Setting the region info in chipset_private */
  IO_VGA_REGION		= s3_ViRGE_io_vga;
  IO_ACCEL_REGION	= s3_ViRGE_io_accel;
  MEM_ACCEL_REGION	= s3_ViRGE_mem_accel;
  MEM_LAW_REGION	= s3_ViRGE_mem_law;
  CHIPSET_INFO		= s3_ViRGE_info;

  /* First try to find a ViRGE. This is done by scanning PCI lists,
   * the card doesn't need to be enabled here yet.
   */	
  if (s3_ViRGE_PCI_detect (dpy)== none){
    NOTICE ("No S3 ViRGE detected. Driver removed. ");
    s3_ViRGE_IO_disable (dpy);
    return -E(CHIPSET,UNKNOWN);
  }
  
  /* We have a card. Now see if this card can be accessed by classic IO.
   * If not, we'll have to try to get the card awake in MMIO mode 
   */
  if (s3_ViRGE_IO_claim (dpy) != -EOK)
    return -EBUSY;
    
  /* When we arrive here, the IO area is free, so IO can be enabled.
   */
  pcicfg_out16(PCI_COMMAND_IO | PCI_COMMAND_MEMORY,
              (pciVirge->bus->number<<24)|(pciVirge->devfn<<16)|PCI_COMMAND);

  s3_unlock(dpy);
  if (s3_ViRGE_setup_memory (dpy) != -EOK) {
    s3_ViRGE_IO_disable (dpy);
    return -EBUSY;
  }
  
  /* Setting some registers */
  s3_defaults_ViRGE( dpy);

  s3_ViRGE_enable_interrupts (dpy);
  return -EOK;
}

void s3_ViRGE_show_hw_pointer (struct kgi_display *dpy, kgi_uint x, kgi_uint y) {
  CRT_out8 (dpy, 0x01, 0x45);
  CRT_out8 (dpy, x & 0xff, 0x47);
  CRT_out8 (dpy, x >> 8, 0x46);
  CRT_out8 (dpy, y & 0xff, 0x49);
  CRT_out8 (dpy, y >> 8, 0x48);
}

void s3_ViRGE_hide_hw_pointer (struct kgi_display *dpy) {
  CRT_out8 (dpy, 0x00, 0x45);
}

void s3_ViRGE_set_origin (struct kgi_display *dpy, kgi_uint x, uint y)
{
  kgiu32 address;
  kgiu32 bpp;
  
	switch(dpy->mode.request.graphtype) {
		case KGIGT_1BIT:
		case KGIGT_4BIT:
		  bpp = 4; break;
		case KGIGT_8BIT:
		  bpp = 8; break;
		case KGIGT_15BIT:
		case KGIGT_16BIT:
		  bpp = 16; break;
		case KGIGT_24BIT:
		  bpp = 24; break;
		case KGIGT_32BIT:
		  bpp = 32; break;
		default:
		  INTERNAL_ERROR;
		  return;
	}
	address=((dpy->mode.request.visible.x*y)+x)*bpp >> 5;
	CRT_out8 (dpy,(address >> 16) & 0x0f,0x69);
	CRT_out8 (dpy,(address >> 8) & 0xff,0x0c);
	CRT_out8 (dpy,address & 0xff,0x0d);
	SPR_out32(dpy,0x1C0, address << 2);

	DEBUG1("ViRGE: Frame origin set to (%d,%d)",x,y);
}

void s3_ViRGE_set_frame (struct kgi_display *dpy, int frame)
{ DEBUG1 ("VIRGE: Set frame"); }

/*
** KGI module interface
*/

int kgim_chipset_init(struct kgi_card *card, struct kgi_display *dpy)
{
	struct chipset_private *private;

	NOTICE(DRIVER_NAME " rev " DRIVER_REV);

	private = kgi_malloc_small(sizeof(*private));
	if (private == NULL) {
		ERROR("ERROR: failed to allocate private data!");
		return -ENOMEM;
	}

	memset(private, 0, sizeof(*private));
	KGIM_PRIVATE(dpy, chipset) = private;
	card->chipset = &CHIPSET_INFO;

	if (s3_ViRGE_init(dpy)) {
		kgi_free_small(private);
		KGIM_PRIVATE(dpy, chipset) = NULL;
		card->chipset = NULL;
		return -E(CHIPSET, UNKNOWN);
	}
	return EOK;
}

void kgim_chipset_done(struct kgi_display *dpy)
{
	GET_PRIVATE;

	s3_ViRGE_disable_interrupts (dpy);
	s3_lock(dpy);

	if (IO_VGA_REGION.baseptr) 
		io_free_region(&IO_VGA_REGION);

	if (IO_ACCEL_REGION.baseptr) 
		io_free_region(&IO_ACCEL_REGION);
	
	if (MEM_ACCEL_REGION.baseptr) 
		mem_free_region(&MEM_ACCEL_REGION);
	
	if (MEM_LAW_REGION.baseptr) 
		mem_free_region(&MEM_LAW_REGION);

	kgi_free_small(private);
	KGIM_PRIVATE(dpy, chipset) = NULL;
	NOTICE(DRIVER_NAME " rev " DRIVER_REV " removed.");
}

static void adjust_timing(struct kgi_mode *mode)
{
	register int foo = (mode->request.graphtype == KGIGT_TEXT16) 
				? mode->request.dpp.x : 8;

	#define ADJUST(X) mode->x.X -= (mode->x.X % foo)

	ADJUST(width);
	ADJUST(blankstart);
	ADJUST(syncstart);
	ADJUST(syncend);
	ADJUST(blankend);

	if (mode->x.total % foo) 
		mode->x.total += foo - (mode->x.total % foo);

	if (((mode->x.syncend / foo) & 0x3F) ==
		((mode->x.syncstart / foo) & 0x3F)) 
		mode->x.syncend -= foo;

	if (((mode->x.blankend / foo) & 0x7F) ==
		((mode->x.blankstart / foo) & 0x7F)) 
		mode->x.blankend -= foo;

	if ((mode->y.syncend & 0x0F) == (mode->y.syncstart & 0x0F)) 
		mode->y.syncend--;

	if ((mode->y.blankend & 0xFF) == (mode->y.blankstart & 0xFF)) 
		mode->y.blankend--;

	#undef ADJUST
}

int kgim_chipset_check_mode (struct kgi_display *dpy, struct kgi_mode *mode,
	enum kgi_tm_cmd *cmd)
{
	GET_PRIVATE;
	kgiu8 bpp = 0;

	if (mode->flags & FL_MON_INTERLACE) {
		NOTICE1("Interlaced modes not supported yet\n");
		return -E(CHIPSET, UNKNOWN);
	}

	bpp = 0;
	switch(mode->request.graphtype) {
		case KGIGT_32BIT:
		  bpp = 32;
		break;

		case KGIGT_24BIT:
		  bpp = 24;
		break;

		case KGIGT_16BIT:
		case KGIGT_15BIT:  /* 15 bit color = KRGB, 16 bits pp */  
		case KGIGT_TEXT16:	
		  bpp = 16;
		  break;

		case KGIGT_8BIT:
		  bpp = 8;
		  break;
		
		case KGIGT_4BIT:
   		  DEBUG1("graphtype %i not supported", mode->request.graphtype);
		  return -E(CHIPSET, UNKNOWN);
		  bpp = 4;
		  break;
		
		case KGIGT_1BIT:
   		  DEBUG1("graphtype %i not supported", mode->request.graphtype);
		  return -E(CHIPSET, UNKNOWN);
	 	  bpp = 4;
		  break;
		
		default:
		DEBUG1("graphtype %i not supported", mode->request.graphtype);
		return -E(CHIPSET, UNKNOWN);
	}

	if (mode->lclk.mul != 1) {
		DEBUG1("Invalid lclk multiplier");
 		return -E(CHIPSET, UNKNOWN);
	}

	switch (mode->lclk.div) 
	{
		case 1:
		case 2:
		case 4:
		case 8:
		case 16:
		case 32:
		break;

		default:
		DEBUG1("invalid CLK to LCLK divider %i", mode->lclk.div);
		return -E(CHIPSET, UNKNOWN);
	}

	switch (*cmd) 
	{
	  case CMD_PROPOSE:
		
		if (mode->request.graphtype == KGIGT_TEXT16) 
		{

			int textx = mode->request.dpp.x;
			int texty = mode->request.dpp.y;

			mode->magnify.x = mode->magnify.y = 1;

			if ((textx < 7) || (9 < textx)) {
				DEBUG1("ERROR: Invalid char width %i", textx);
				return -EINVAL;
			}

			if (!(XVIRTUAL || YVIRTUAL)) {
				DEBUG1("ERROR: Need x or y virtual!");
				return -EINVAL;
			}

			if (!XVIRTUAL) {
				XVIRTUAL = TEXT_MEMORY_SIZE * textx / 
					(YVIRTUAL / texty);
				DEBUG1("Setting xvirtual = %i", XVIRTUAL);
			}

			if (!YVIRTUAL) {
				YVIRTUAL = TEXT_MEMORY_SIZE * texty / 
					(XVIRTUAL / textx);
				DEBUG1("Setting yvirtual = %i", YVIRTUAL);
			}

			if (((XVIRTUAL / textx) * (YVIRTUAL / texty)) > 
			    TEXT_MEMORY_SIZE) {
				ERROR("ERROR: Not enough text memory");
				return -ENOMEM;
			}

			mode->request.visible.x -=
				(mode->request.visible.x % textx);
			mode->request.visible.y -=
				(mode->request.visible.y % texty);

		} else /* mode->request.graphtype == KGIGT_TEXT16 */
		  {
			mode->magnify.x = mode->magnify.y = 1;

			while ((XVISIBLE * mode->magnify.x) < 640) 
			  mode->magnify.x++;

			while ((YVISIBLE * mode->magnify.y) < 350) 
			  mode->magnify.y++;
			
			if ((mode->magnify.y > 2) || (mode->magnify.x > 2))
			  return -EINVAL;

			if (XVIRTUAL <= 300) XVIRTUAL = 300;
			else if (XVIRTUAL <= 320) XVIRTUAL = 320;
			else if (XVIRTUAL <= 640) XVIRTUAL = 640;
			else if (XVIRTUAL <= 800) XVIRTUAL = 800;
			else if (XVIRTUAL <= 1024) XVIRTUAL = 1024;
			else if (XVIRTUAL <= 1152) XVIRTUAL = 1152;
			else if (XVIRTUAL <= 1280) XVIRTUAL = 1280;
			else if (XVIRTUAL <= 1600) XVIRTUAL = 1600;
			else if (XVIRTUAL <= 2048) XVIRTUAL = 2048;
			else {
			  ERROR("Too many horizontal pixels (%i)",
				XVIRTUAL);
			  return -E(CHIPSET, UNKNOWN);
			}
			if ((XVIRTUAL*YVIRTUAL*bpp/8) > CHIPSET_INFO.ramsize) {
				ERROR("Not enough memory for %ix%i@%ibpp",
					XVIRTUAL, YVIRTUAL, bpp);
				return -ENOMEM;
			}

			mode->request.visible.x &= 0x7FFFFFF8;
			mode->request.visible.y &= 0x7FFFFFF8;
		  }

		mode->x.width = mode->request.visible.x * mode->magnify.x;
		mode->y.width = mode->request.visible.y * mode->magnify.y;
		return EOK;

	  case CMD_LOWER:

		if (mode->dclk < CHIPSET_INFO.dclk.min) {
			ERROR("ERROR: DCLK too low (%i)\n", mode->dclk);
			return -E(CHIPSET, UNKNOWN);
		}

		mode->dclk = (mode->dclk > CHIPSET_INFO.dclk.max)
			? CHIPSET_INFO.dclk.max : mode->dclk;

		adjust_timing(mode);
		return EOK;

	  case CMD_RAISE:

		if (mode->dclk > CHIPSET_INFO.dclk.max) {
			ERROR("ERROR: DCLK too high (%i)\n", mode->dclk);
			return -E(CHIPSET, UNKNOWN);
		}

		mode->dclk = (mode->dclk < CHIPSET_INFO.dclk.min)
			? CHIPSET_INFO.dclk.min : mode->dclk;

		adjust_timing(mode);
		return EOK;

	  case CMD_CHECK:
		if (mode->request.graphtype == KGIGT_TEXT16) {
			if ((mode->magnify.x != 1) || 
			    (mode->magnify.y != 1) || 
			    (mode->request.dpp.x < 7) ||
			    (mode->request.dpp.x > 9) ||
			    (mode->request.dpp.y < 1) ||
			    (mode->request.dpp.y > 32) ||
			    (((XVIRTUAL/mode->request.dpp.x) *
			      (YVIRTUAL/mode->request.dpp.y)) > TEXT_MEMORY_SIZE) ||
			    (mode->dclk > CHIPSET_INFO.dclk.max) ||
			    (mode->dclk < CHIPSET_INFO.dclk.min)) {
				return -E(CHIPSET, UNKNOWN);
			}

			mode->flags |= FL_CHIP_COLORTEXT;
			mode->fb = MEM_LAW_REGION.baseptr;
			mode->fb_size = TEXT_MEMORY_SIZE;
			mode->fsize = (8 * mode->request.virt.x *
				mode->request.virt.y) /
				(mode->request.dpp.x * mode->request.dpp.y);
			mode->io = &text16_color_io;
			mode->mem = &text16_color_mem;

		} else /* mode->request.graphtype == KGIGT_TEXT16 */
		{
			switch (XVIRTUAL) {
				case 300:
				case 320:
				case 640:
				case 800:
				case 1024:
				case 1152:
				case 1280:
				case 1600:
				case 2048:
				  break;
				
				default:
				  return -E(CHIPSET, UNKNOWN);
			}

			if (((XVIRTUAL * YVIRTUAL * bpp / 8) >
			     CHIPSET_INFO.ramsize) ||
			    (mode->clk.mul != 1) ||
			    (mode->clk.div != 1) ||
			    (mode->dclk < CHIPSET_INFO.dclk.min) ||
			    (mode->dclk > CHIPSET_INFO.dclk.max)) {
				return -E(CHIPSET, UNKNOWN);
			}

			mode->fb	= MEM_LAW_REGION.baseptr;
			mode->fb_size	= CHIPSET_INFO.ramsize;
			mode->fsize	= mode->request.virt.x *
					  mode->request.virt.y *
					  bpp / 8;
			mode->io	= NULL;
			mode->mem	= NULL;
			mode->mmio	= NULL;
		  }

	      #define TM(X) (((unsigned) CHIPSET_INFO.maxdots.x < mode->x.X)\
	      || ((unsigned) CHIPSET_INFO.maxdots.y < mode->y.X))

		if (TM(width) || TM(blankstart) || TM(syncstart) || 
		    TM(syncend) || TM(blankend) || TM(total)) 
		{
			return -E(CHIPSET, UNKNOWN);
		}
              #undef TM

		adjust_timing(mode);
		mode->mmio = & private->lram;

		return EOK;

	  default:
		INTERNAL_ERROR;
		return -E(CHIPSET, UNKNOWN);
	}

#undef XVIRTUAL
#undef YVIRTUAL
}

void kgim_chipset_set_mode (struct kgi_display *dpy, struct kgi_mode *mode)
{
	IS_CRITICAL;
	GET_PRIVATE;
	struct statetable setup = default_state;

	register kgiu32 foo;
	kgiu32 DCLKdiv = 0;
	kgiu8 crt11;
	DEBUG1("Setting chispet for : (%d,%d)[%d,%d]*(%d,%d)*(%d,%d) (mode %d)",
	  mode->x.width, mode->y.width, mode->request.virt.x, mode->request.virt.y,
	  mode->request.dpp.x, mode->request.dpp.y, mode->magnify.x, mode->magnify.y, mode->request.graphtype);
	/* Set up bits per pixel and logical screen width */
	/* foo has to contain the virtual line width in QUADWORDS (bytes / 8) */
	switch(mode->request.graphtype) {
		case KGIGT_TEXT16:
		  foo = (2 * mode->request.virt.x / mode->request.dpp.x) / 4;
		  break;
		case KGIGT_1BIT:
		  foo = mode->request.virt.x >>	5;
		  setup.MISC &= ~0x01;  
		  setup.CRT[0x66] &= ~0x01;
		  break;
		case KGIGT_4BIT:
		  foo = mode->request.virt.x >> 4;
		  setup.CRT[0x3A] &= ~CR3A_NO4BPPMODE;
		  setup.CRT[0x66] &= ~0x01;
		  break;
		case KGIGT_8BIT:
		  setup.CRT[0x3A] |= CR3A_NO4BPPMODE;
		  setup.CRT[0x66] |= 0x01;
		  foo = mode->request.virt.x >> 3;
		  break;
		case KGIGT_15BIT:
		case KGIGT_16BIT:
		  setup.CRT[0x3A] |= CR3A_NO4BPPMODE;
		  setup.CRT[0x66] |= 0x01;
		  setup.CRT[0x43] |= 0x80;
		  foo = mode->request.virt.x >> 2;
		  break;
		case KGIGT_24BIT:
		  setup.CRT[0x3A] |= CR3A_NO4BPPMODE;
		  setup.CRT[0x66] |= 0x01;
		  foo = mode->request.virt.x * 3 / 8;
		  break;
		case KGIGT_32BIT:
		  setup.CRT[0x3A] |= CR3A_NO4BPPMODE;
		  setup.CRT[0x66] |= 0x01;
		  foo = mode->request.virt.x >> 1;
		  break;
		default:
		  INTERNAL_ERROR;
		  return;
	}

	setup.CRT[CRT_LOGICALWIDTH] = foo & 0xFF;
	setup.CRT[0x51] |= (foo & 0x100) ? CR51_LOGICALWIDTH_B8 : 0;
	setup.CRT[0x51] |= (foo & 0x200) ? CR51_LOGICALWIDTH_B9 : 0;

	if (mode->request.graphtype == KGIGT_TEXT16) {
		setup.CRT[0x17] = CR17_CGA_ADDR_WRAP | CR17_WORDMODE |
			CR17_ENABLE_SYNC | CR17_CGA_BANKING | 
			CR17_HGC_BANKING;
		setup.CRT[0x31] |= CR31_HISPEEDTXT;
		setup.CRT[0x34] |= CR34_ENB_FIFO; 

		setup.SEQ[0x02] = SR02_PLANE0 | SR02_PLANE1;

		setup.GRC[0x05] |= GR5_ODD_EVEN;
		setup.GRC[0x06] |= GR6_MAP_B8_32 | GR6_CHAIN_OE;
	
		switch((DCLKdiv = mode->request.dpp.x)) {
			case 7: 
			  setup.CRT[0x32] |= 0x01; 
			  break;
			case 8: 
			  setup.SEQ[0x01] |= SR01_8DOT_CHCLK; 
			  break;
			case 9: 
			  break;
			default:
			  INTERNAL_ERROR;
			  return;
		}

		setup.CRT[0x09] |= mode->request.dpp.y - 1;
		setup.ATC[0x13] = (mode->request.dpp.x == 9) ? 0x08 : 0x00;
		setup.CRT[CRT_MEMCONTROL2] = 8;
		setup.CRT[CRT_MEMCONTROL3] = 7;
		setup.CRT[0x62] = 84;

	} else {
		DCLKdiv = 8;
		setup.CRT[0x17] = CR17_CGA_ADDR_WRAP | CR17_WORDMODE |
			CR17_ENABLE_SYNC | CR17_CGA_BANKING | 
			CR17_HGC_BANKING;
		
		setup.SEQ[0x01] |= SR01_8DOT_CHCLK;
		setup.SEQ[0x02] |= SR02_PLANE_MASK;
		setup.ATC[0x10] |= AR10_GRAPHICS;
		setup.GRC[0x05] |= GR5_8BIT_COLOR;
		setup.GRC[0x06] |= GR6_GRAPHMODE | GR6_MAP_A0_64;
		setup.CRT[0x31] |= CR31_ENHANCED;
		setup.CRT[0x58] |= CR58_LAW_ENABLE;
  		setup.CRT[0x09] |= (mode->flags & FL_CHIP_DOUBLESCAN)
			? CR09_DOUBLESCAN : 0;
		
		if (mode->request.virt.x == 2048)
		  setup.CRT[0x31] |= CR31_2PAGE_SCRN;

		setup.CRT[CRT_MEMCONTROL2] = mode->chipset;
		setup.CRT[CRT_MEMCONTROL3] = 0xFF;
	}	

	/* Set up horizontal timing */

	/* Horizontal display end */
	foo = (mode->x.width / DCLKdiv) - 1;
	setup.CRT[0x01] |= foo & 0xFF;
	setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HDISPLAYEND_B8 : 0;

	/* Horizontal blank start */ 
	foo = (mode->x.blankstart / DCLKdiv);
	setup.CRT[0x02] |= foo & 0xFF;
	setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HBLANKSTART_B8 : 0;

	/* Horizontal sync start */ 
	foo = (mode->x.syncstart / DCLKdiv);
	setup.CRT[0x04] |= foo & 0xFF;
	setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HSYNCSTART_B8 : 0;

	/* Horizontal sync end */
	foo = (mode->x.syncend / DCLKdiv);
	setup.CRT[0x05] |= foo & CR05_HSE_MASK;
	setup.CRT[0x5D] |= ((foo-(mode->x.syncstart / DCLKdiv)) > 0x20)
		? CR5D_HSYNCEND_B5 : 0;

	/* Horizontal blank end */
	foo = (mode->x.blankend / DCLKdiv);
	setup.CRT[0x03] |= foo & CR03_HBE_MASK;
	setup.CRT[0x05] |= (foo & 0x20) ? CR05_HBLANKEND_B5 : 0;
	setup.CRT[0x5D] |= ((foo-(mode->x.blankstart / DCLKdiv)) > 0x40)
		? CR5D_HBLANKEND_B6 : 0;

	/* Horizontal total */
	foo = (mode->x.total / DCLKdiv)- 5;
	setup.CRT[0x00] |= foo & 0xFF;
	setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HTOTAL_B8 : 0;

	/* Set up vertical timing */

	/* Vertical display end */
	foo = mode->y.width - 1;
	setup.CRT[0x12] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VDISPLAYEND_B8  : 0;
	setup.CRT[0x07] |= (foo & 0x200) ? CR07_VDISPLAYEND_B9  : 0;
	setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VDISPLAYEND_B10 : 0;

	/* Vertical blank start */
	foo = mode->y.blankstart - 1;
	setup.CRT[0x15] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VBLANKSTART_B8  : 0;
	setup.CRT[0x09] |= (foo & 0x200) ? CR09_VBLANKSTART_B9  : 0;
	setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VBLANKSTART_B10 : 0;

	/* Vertical sync start */ 
	foo = mode->y.syncstart;
	setup.CRT[0x10] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VSYNCSTART_B8  : 0;
	setup.CRT[0x07] |= (foo & 0x200) ? CR07_VSYNCSTART_B9  : 0;
	setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VSYNCSTART_B10 : 0;

	/* Vertical sync end */
	foo = mode->y.syncend;
	setup.CRT[0x11] |= foo & CR11_VSYNCEND_MASK;

	/* Vertical blank end */
	foo = mode->y.blankend - 1;
	setup.CRT[0x16] |= foo & 0xFF;

	/* Vertical total */
	foo = mode->y.total - 2;
	setup.CRT[0x06] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VTOTAL_B8  : 0;
	setup.CRT[0x07] |= (foo & 0x200) ? CR07_VTOTAL_B9  : 0;
	setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VTOTAL_B10 : 0;

	/* Setup interlace modes */
	setup.CRT[0x42] |= ((mode->flags & FL_MON_INTERLACE)
	 		  ? CR42_INTERLACED
	 		  : 0);
	setup.CRT[0x3C] = mode->y.total >> 1;
	
	/* Setup synchronisation pulses polarity */
	setup.MISC |= (mode->x.polarity > 0) ? 0 : MISC_NEG_HSYNC;
	setup.MISC |= (mode->y.polarity > 0) ? 0 : MISC_NEG_VSYNC;

	/* Set up the FIFO control parameters */
	foo = (mode->x.total / DCLKdiv)-10;
	setup.CRT[CRT_FIFOSTART] |= foo & 0xFF;	/* Start display FIFO fetch */
	setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_FIFOSTART_B8 : 0;

	#if DEBUG_LEVEL > 1
        	NOTICE("This setup is about to be programmed\n");
	        dump_state(dpy,&setup);
	#endif
	/* Ok.  We'll do the actual hardware programming now. */ 
	ENTER_CRITICAL;

	/* Switch off interrupts */
	if (s3_ViRGE_interrupts.enabled) {
	  /* Switch off ViRGE interrupt generator and vertical retrace */
	  CRT_out8(dpy, (CRT_in8(dpy, 0x11) & 0xDF) | 0x10, 0x11);
  	  /* Disable all enhanced interrupts. */
  	  mem_out32(0x00000000, MEM_VGA + 0X0504);
	}

	/* Switch off Streams Processor */
	for (foo = 100000; foo > 0; foo--) 
    	  if (mem_in8(MEM_VGA + 0x03DA) & 0x08) break;
	if (foo == 0) DEBUG1 ("ViRGE: VSYNC not found.");
  	SPR_out32(dpy,0x0200,0x0000C000);
  	CRT_out8 (dpy, CRT_in8 (dpy, 0x67) & ~0x0C, 0x67); 
	
	/* Wait for S3d Engine */
	for (foo = 100000; foo > 0; foo-- ) 
	  if (mem_in32(MEM_VGA + 0X0504) & 0x00002000) break;
	if (foo == 0) DEBUG1 ("ViRGE: S3d Engine remains busy.");
	/* Reset S3d Engine */
  	mem_out32(0x00008000, MEM_VGA + 0X0504);
  	mem_out32(0x00004000, MEM_VGA + 0X0504);
	/* Wait till S3d Engine is ready. */
	for (foo = 100000; foo > 0; foo-- ) 
	  if (mem_in32(MEM_VGA + 0X0504) & 0x00002000) break;
	if (foo == 0) DEBUG1 ("ViRGE: S3D Engine reset failed.");

	/* Switch off sync signals */  
	CRT_out8(dpy, CRT_in8(dpy, CRT_EXTSYNC1) |0x06, CRT_EXTSYNC1);
	/* Switch off display */
	setup.SEQ[0x01] |= SR01_DISPLAY_OFF; 
	SEQ_out8(dpy, setup.SEQ[0x01], 0x01);

	/* Prepare subsystems */
	kgim_monitor_set_mode(dpy, mode);
	NOTICE2("monitor set");
	kgim_ramdac_set_mode(dpy, mode);
	NOTICE2("ramdac set");
	kgim_clock_set_mode(dpy, mode);
	NOTICE2("clock set");
	restore_state(dpy, &setup);
	if (cardtype == dxgx || cardtype == gx2) {
	  CRT_out8 (dpy, 0x80, 0x86); /* Disable DAC power savings */
	  CRT_out8 (dpy, 0x00, 0x90); /* Disable Stream display fetch length control */
	}
	if (cardtype == vx) {
	  CRT_out8 (dpy, 0x00, 0x50);
	  CRT_out8 (dpy, setup.CRT[0x66],0x63); /* Somehow this register moved :) */
	}
	NOTICE2("chip set");
	kgim_accel_set_mode (dpy, mode);
	NOTICE2("accelerator set");

	/* Turn on sync signals */
	CRT_out8(dpy, CRT_in8(dpy, CRT_EXTSYNC1) & 0xf9, CRT_EXTSYNC1);
	/* Enable screen */
	setup.SEQ[0x01] &= ~SR01_DISPLAY_OFF; 
	SEQ_out8(dpy, setup.SEQ[0x01], 0x01);
        mem_out8(0x20,MEM_VGA+0x03c0);
	mem_out8(0x01,MEM_VGA+0x03c3);

	if (s3_ViRGE_interrupts.enabled) {
  	  /* Clear vertical retrace interrupt and enable it. */
  	  if (s3_ViRGE_interrupts.bitmap & 0x00000100) {
   	    crt11 = CRT_in8(dpy,0x11);
  	    CRT_out8(dpy,crt11 & ~0x30,0x11);  
  	    CRT_out8(dpy,crt11 | 0x10,0x11);
  	  }  
  
  	  /* Clear all pending enhanced interrupts and enable them */
  	  mem_out32(0x0000007F, MEM_VGA + 0X0504);
  	  mem_out32(s3_ViRGE_interrupts.bitmap, MEM_VGA + 0X0504);
	}
	
	LEAVE_CRITICAL;
	
	memcpy(&(dpy->mode), mode, sizeof(dpy->mode));

	private->split = -1;
	private->org_xdot = private->org_ydot = private->org_offs = -1;

	if (mode->request.graphtype == KGIGT_TEXT16) {
		dpy->set_font	= s3_text_set_font;
		dpy->set_split	= s3_set_split;
		dpy->set_frame	= s3_ViRGE_set_frame;
		dpy->set_origin	= s3_text_set_origin;

		/* Provide hardware cursor in text mode */
		if (HARDWARE_TEXT_CURSOR) {
			dpy->show_cursor = s3_show_hw_text_cursor;
			dpy->hide_cursor = s3_hide_hw_text_cursor;
			dpy->undo_cursor = NULL;
		} else {
			dpy->show_cursor = s3_show_sw_text_cursor;
			dpy->hide_cursor = s3_hide_sw_text_cursor;
			dpy->undo_cursor = s3_undo_sw_text_cursor;
		}

		/* Provide software cursor too in case DAC didn't already */
		if (!dpy->show_pointer) {
			dpy->show_pointer = s3_show_sw_text_pointer;
			dpy->hide_pointer = s3_hide_sw_text_pointer;
			dpy->undo_pointer = s3_undo_sw_text_pointer;
		}

		dpy->fb.size.x = mode->request.virt.x / mode->request.dpp.x;
		dpy->fb.size.y = mode->request.virt.y / mode->request.dpp.y;
		dpy->fb.bpp    = MAP_16_BPP;
		dpy->fb.pbuf   = MEM_LAW;
		dpy->fb.zbuf   = NULL;
		dpy->fb.ops    = &s3_text16;

		private->fb_size = dpy->fb.size.x * dpy->fb.size.y;

	} else {
	  if (!dpy->show_pointer) {
	    dpy->show_pointer = s3_ViRGE_show_hw_pointer;
	    dpy->hide_pointer = s3_ViRGE_hide_hw_pointer;
	    dpy->undo_pointer = NULL;
	  }
	  dpy->fb.ops = NULL;
	  dpy->set_font = NULL;
	  dpy->set_split = s3_set_split;
	  dpy->set_origin = s3_ViRGE_set_origin;
	  dpy->set_frame = NULL; /*s3_ViRGE_set_frame*/;
	}

	NOTICE1("mode set");
}

int kgim_chipset_command(struct kgi_graphic *graph, kgi_uint cmd, kgi_uint arg)
{
  kgi_coord *coord;
  DEBUG1("chipset_command: %d",cmd);
  switch (cmd) {
    case CHIP_SETVISFRAME:
		coord = (kgi_coord *)(graph->io_buf);
		if (coord->x < 0 || coord->x > graph->dev.mode.virt.x)
			return -E(CHIPSET,BOGUS);
		
		if (coord->y < 0 || coord->y > graph->dev.mode.virt.y)
			return -E(CHIPSET,BOGUS);
		
		kgi_set_origin(&(graph->dev), 0, coord->x, coord->y);
		
		return 0;
    default:
      return -E(CHIPSET, UNKNOWN);
  }
}
